import operator, os, warnings
from collections.abc import Iterable
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
import astropy.units as u
from astropy.nddata import CCDData, StdDevUncertainty
# specutils
from specutils import Spectrum1D
# drpy
from drpy import conf

# Configurations
unit_ccddata = u.Unit(conf.unit_ccddata)
unit_spectral_axis = u.Unit(conf.unit_spectral_axis)
unit_flux = u.Unit(conf.unit_flux)

__all__ = ['_validateBool', '_validateString', '_validateRange', '_validateInteger', 
           '_validateLength', '_validate1DArray', '_validateNDArray', 
           '_validateCCDData', '_validateSpectrum1D', '_validateCCDList', 
           '_validateCCD', '_validateSpectrum', '_validateBins', '_validateAperture', 
           '_validatePath']
# Types

def _validateBool(arg, name):
    """Validate a `bool` object."""

    if not isinstance(arg, (bool, np.bool_)):
        raise TypeError(
            'Invalid type `{}` for ``{}``. `bool` is expected.'
            .format(type(arg), name))

    return True


def _validateString(arg, name, value_list=None):
    """Validate a `str` object."""

    if not isinstance(arg, (str, np.str_)):
        raise TypeError(
            'Invalid type `{}` for ``{}``. `str` is expected.'.format(type(arg), name))

    if value_list is not None:
        if arg not in value_list:
            raise ValueError('``{}`` should be in `{}`.'.format(name, value_list))

    return True


# With properties

def _validateRange(arg, name, value_range=(None, None), is_closed=(None, None)):
    """Validate whether ``arg`` is in the specified range."""

    if (value_range[0] is not None) & (value_range[1] is not None):
        if is_closed[0] & is_closed[1]:
            if (arg < value_range[0]) | (value_range[1] < arg):
                raise ValueError(
                    '{} must be in range [{}, {}]. {} is given.'.format(
                        name, value_range[0], value_range[1], arg)
                )
        elif is_closed[0]:
            if (arg < value_range[0]) | (value_range[1] <= arg):
                raise ValueError(
                    '{} must be in range [{}, {}). {} is given.'.format(
                        name, value_range[0], value_range[1], arg)
                )
        elif is_closed[1]:
            if (arg <= value_range[0]) | (value_range[1] < arg):
                raise ValueError(
                    '{} must be in range ({}, {}]. {} is given.'.format(
                        name, value_range[0], value_range[1], arg)
                )
        else:
            if (arg <= value_range[0]) | (value_range[1] <= arg):
                raise ValueError(
                    '{} must be in range ({}, {}). {} is given.'.format(
                        name, value_range[0], value_range[1], arg)
                )
    elif value_range[0] is not None:
        if is_closed[0]:
            if arg < value_range[0]:
                raise ValueError(
                    '{} must be in range [{}, inf). {} is given.'.format(
                        name, value_range[0], arg)
                )
        else:
            if arg <= value_range[0]:
                raise ValueError(
                    '{} must be in range ({}, inf). {} is given.'.format(
                        name, value_range[0], arg)
                )
    elif value_range[1] is not None:
        if is_closed[1]:
            if arg > value_range[1]:
                raise ValueError(
                    '{} must be in range (-inf, {}]. {} is given.'.format(
                        name, value_range[1], arg)
                )
        else:
            if arg >= value_range[1]:
                raise ValueError(
                    '{} must be in range (-inf, {}). {} is given.'.format(
                        name, value_range[1], arg)
                )

    return True


def _validateInteger(arg, name, value_range=(None, None), is_closed=(None, None)):
    """Validate an `int` object."""

    try:
        arg = operator.index(arg)

    except TypeError:
        raise TypeError(
            'Invalid type `{}` for ``{}``. `int` is expected.'.format(type(arg), name))

    _validateRange(arg, name, value_range, is_closed)

    return True


def _validateLength(arg, name, length):
    """Validate length of an array."""

    narg = np.array(arg)

    if narg.shape[0] < length:
        raise ValueError(
            '``{}`` should be of length {}. {} is given.'.format(
            name, length, narg.shape[0]))

    elif narg.shape[0] > length:
        narg = narg[:length]
    
    return narg


def _validate1DArray(arg, name, length, accept_scalar):
    """Validate a 1-dimensional array."""

    if np.ndim(arg) == 0:
        if accept_scalar:
            narg = np.zeros(length) + arg

        else:
            raise ValueError('``{}`` must be 1-dimensional.'.format(name))

    elif np.ndim(arg) == 1:
        narg = _validateLength(arg, name, length)
    
    else:
        raise ValueError('``{}`` must be 1-dimensional, when an array.'.format(name))
    
    return narg


def _validateNDArray(arg, name, ndim):
    """Validate a 2-dimensional array."""

    if np.ndim(arg) == ndim:
        narg = np.array(arg)

    else:
        raise ValueError(
            '``{}`` must be {}-dimensional.'.format(name, ndim))

    return narg


# Specific arguments

def _validateCCDData(ccd, name):
    """Validate a CCD image."""

    if not isinstance(ccd, (np.ndarray, CCDData)):
        raise TypeError('Invalid type `{}` for ``{}``.'.format(type(ccd), name))
    # May have (or not have) uncertainty frame or mask frame.
    elif isinstance(ccd, CCDData):
        nccd = ccd.copy()
    # Do not have uncertainty frame. May have (or not have) mask frame.
    else:
        nccd = CCDData(deepcopy(ccd), unit=unit_ccddata)

    return nccd


def _validateSpectrum1D(spectrum, name):
    """Validate a spectrum."""

    if not isinstance(spectrum, (np.ndarray, Spectrum1D)):
        raise TypeError('Invalid type `{}` for ``{}``.'.format(type(spectrum), name))
    
    elif isinstance(spectrum, Spectrum1D):
        new_spectrum = deepcopy(spectrum)
    
    else:

        if spectrum.ndim == 1:

            if np.ma.isMaskedArray(spectrum):
                flux = spectrum.data * unit_flux
                mask = spectrum.mask

            else:
                flux = spectrum * unit_flux
                mask = None

            new_spectrum = Spectrum1D(flux=flux, mask=mask)

        elif spectrum.ndim == 2:

            n_spec = spectrum.shape[0]

            if np.ma.isMaskedArray(spectrum):

                mask = spectrum.mask[:3].any(axis=0)

                # flux
                if n_spec == 1:
                    spectral_axis = np.arange(spectrum.shape[1]) * unit_spectral_axis
                    flux = spectrum.data[0] * unit_flux
                    uncertainty = None

                # spectral axis, flux
                elif n_spec == 2:
                    spectral_axis = spectrum.data[0] * unit_spectral_axis
                    flux = spectrum.data[1] * unit_flux
                    uncertainty = None

                # spectral axis, flux, uncertainty
                else:
                    spectral_axis = spectrum.data[0] * unit_spectral_axis
                    flux = spectrum.data[1] * unit_flux
                    uncertainty = StdDevUncertainty(spectrum.data[2])

            else:

                mask = None

                # flux
                if n_spec == 1:
                    spectral_axis = np.arange(spectrum.shape[1]) * unit_spectral_axis
                    flux = spectrum[0] * unit_flux
                    uncertainty = None

                # spectral axis, flux
                elif n_spec == 2:
                    spectral_axis = spectrum[0] * unit_spectral_axis
                    flux = spectrum[1] * unit_flux
                    uncertainty = None

                # spectral axis, flux, uncertainty
                else:
                    spectral_axis = spectrum[0] * unit_spectral_axis
                    flux = spectrum[1] * unit_flux
                    uncertainty = StdDevUncertainty(spectrum[2])

            new_spectrum = Spectrum1D(
                spectral_axis=spectral_axis, flux=flux, uncertainty=uncertainty, 
                mask=mask)

        else:
            raise ValueError(
                '``{}`` must be 1 or 2-dimensional, when an array.'.format(name))
    
    return new_spectrum


def _validateCCDList(ccdlist, n_frame, n_dim):
    """Validate a list of frames."""
    
    # Type
    if isinstance(ccdlist, Iterable):

        # Length
        if len(ccdlist) < n_frame:
            raise ValueError(
                f'``ccdlist`` should contain at least {n_frame} frames.')

        # Type and dimension of elements
        nccdlist = list()
        for ccd in ccdlist:
            nccd = _validateCCDData(ccd, 'ccd')
            if nccd.ndim != n_dim:
                raise ValueError(
                    f'A dimension of `{n_dim}` is expected. `{nccd.ndim}` is given.')
            else:
                nccdlist.append(nccd)

    else:
        raise TypeError(f'Invalid type `{type(ccdlist)}` for ``ccdlist``.')
    
    return nccdlist


def _validateCCD(ccd, name, use_uncertainty, use_mask, transpose):
    """Validate a CCD image."""
    
    nccd = _validateCCDData(ccd, name)

    # Data
    data_arr = deepcopy(nccd.data)
    
    # Uncertainty
    if use_uncertainty:
        if nccd.uncertainty is None:
            warnings.warn(
                'The input uncertainty is unavailable. All set to zero.', 
                RuntimeWarning)
            uncertainty_arr = np.zeros_like(data_arr)
        else:
            uncertainty_arr = deepcopy(nccd.uncertainty.array)
    else:
        uncertainty_arr = np.zeros_like(data_arr)
    
    # Mask
    if use_mask:
        if nccd.mask is None:
            warnings.warn(
                'The input mask is unavailable. All set unmasked.', RuntimeWarning)
            mask_arr = np.zeros_like(data_arr, dtype=bool)
        else:
            mask_arr = deepcopy(nccd.mask)
    else:
        mask_arr = np.zeros_like(data_arr, dtype=bool)

    if transpose:
        data_arr = data_arr.T
        uncertainty_arr = uncertainty_arr.T
        mask_arr = mask_arr.T
    
    return nccd, data_arr, uncertainty_arr, mask_arr


def _validateSpectrum(spectrum, name, use_uncertainty, use_mask):
    """Validate a spectrum."""

    new_spectrum = _validateSpectrum1D(spectrum, name)
    
    # Data
    spectral_axis = deepcopy(new_spectrum.spectral_axis.value)
    flux = deepcopy(new_spectrum.flux.value)
    
    # Uncertainty
    if use_uncertainty:
        if new_spectrum.uncertainty is None:
            warnings.warn(
                'The input uncertainty is unavailable. All set to zero.', 
                RuntimeWarning)
            uncertainty = np.zeros_like(flux)
        else:
            uncertainty = deepcopy(new_spectrum.uncertainty.array)
    else:
        uncertainty = np.zeros_like(flux)
    
    # Mask
    if use_mask:
        if new_spectrum.mask is None:
            warnings.warn(
                'The input mask is unavailable. All set unmasked.', RuntimeWarning)
            mask = np.zeros_like(flux, dtype=bool)
        else:
            mask = deepcopy(new_spectrum.mask)
    else:
        mask = np.zeros_like(flux, dtype=bool)
    
    return new_spectrum, spectral_axis, flux, uncertainty, mask


def _validateBins(bins, length):
    """Validate bins and derive bin edges."""

    if np.ndim(bins) == 0:
        # Validate
        _validateInteger(bins, 'bins', (1, length), (True, True))
        # Generate bin edges
        bin_edges = np.linspace(0, length, bins + 1).astype(int)

    elif np.ndim(bins) == 1:
        # Generate bin edges
        bin_edges = np.array(bins).astype(int)
        # Validate
        _validateInteger(bin_edges.min(), 'bin_edges', (0, length), (True, True))
        _validateInteger(bin_edges.max(), 'bin_edges', (0, length), (True, True))
        if np.any(bin_edges[:-1] > bin_edges[1:]):
            raise ValueError('``bins`` must increase monotonically, when an array')

    else:
        raise ValueError('``bins`` must be 1-dimensional, when an array')

    bin_edges = np.vstack([bin_edges[:-1], bin_edges[1:]]).T

    return bin_edges


def _validateAperture(aper_width):
    """Validate aperture width."""

    if np.ndim(aper_width) == 0:

        _validateRange(aper_width, 'aper_width', (0, None), (False, None))

        aper_width = _validate1DArray(aper_width, 'aper_width', 2, True) / 2

    elif np.ndim(aper_width) == 1:

        aper_width = _validate1DArray(aper_width, 'aper_width', 2, True)

        _validateRange(
            aper_width.sum(), 'aper_width.sum()', (0, None), (False, None))

    else:
        raise ValueError('``aper_width`` must be 1-dimensional, when an array')

    return aper_width


def _validatePath(save, path, title):
    """Validate path arguments."""

    _validateBool(save, 'save')

    if save:

        if path is None:
            path = os.getcwd()

        else:
            _validateString(path, 'path')

        if isinstance(title, str):
            fig_name = f'{title}.png'.replace(' ', '_')
            fig_path = os.path.join(path, fig_name)
        
        else:
            fig_path = list()
            for t in title:
                fig_name = f'{t}.png'.replace(' ', '_')
                fig_path.append(os.path.join(path, fig_name))

    else:
        fig_path = None

    return fig_path


# NumPy
import numpy as np
# Astropy
from astropy.stats import sigma_clip
# SciPy
from scipy import interpolate
from scipy.ndimage import gaussian_filter
# drpy
from drpy.validate import _validateBool, _validateInteger, _validate1DArray

__all__ = ['Poly1D', 'Spline1D', 'Spline2D', 'GaussianSmoothing2D']


def Poly1D(x, y, weight=None, mask=None, degree=1, n_iter=5, sigma_lower=None, 
           sigma_upper=None, grow=False, use_relative=False):
    """Fit 1-dimensional polynomial function.
    
    Sigma clipping is used to mask bad data points.

    Parameters
    ----------
    x : array_like
        Input dimension of data points – must be increasing.

    y : array_like
        Input dimension of data points.

    weight : array_like, optional
        Weights for polynomial fitting. Must be positive. If `None` (default), weights 
        are all equal.

    mask : array_like, optional
        Initial mask for polynomial fitting. If `None` (default), data points are all 
        unmasked.

    degree : int, optional
        Degree of the fitting polynomial.
        Default is `1`.

    n_iter : int, optional
        Number of sigma slipping iterations. Must be ``n_iter`` >= `0`.
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    grow : scalar or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits. 

    use_relative : bool, optional
        If `True`, relative residual is used in sigma slipping instead of residual. 
        Default is `False`.

    Returns
    -------
    p : callable
        Callable object returned by `~numpy.poly1d`.

    residual : `~numpy.ndarray`
        Residual or relative residual, depending on ``use_relative``.

    threshold_lower : scalar or `None`
        Lower threshold for clipping.

    threshold_upper : scalar or `None`
        Upper threshold for clipping.

    master_mask : `~numpy.ndarray`
        Final mask after clipping.
    """

    if weight is None:
        weight = np.ones_like(y)

    if mask is None:
        mask = np.zeros_like(y, dtype=bool)

    _validateInteger(n_iter, 'n_iter', (0, None), (True, None))

    _validateBool(use_relative, 'use_relative')

    # Polynomial fitting iteratively (!!! only ``n_iter`` effective iterations !!!)
    master_mask = mask & True
    for k in range(n_iter + 1):

        p = np.poly1d(
            np.polyfit(x[~master_mask], y[~master_mask], degree, w=weight[~master_mask])
        )

        y_fit = p(x)

        # Residual
        residual = y - y_fit
        if use_relative:
            residual /= y_fit

        # Sigma clipping
        if n_iter == 0:
            threshold_lower, threshold_upper = None, None

        elif k < n_iter:
            residual[mask] = np.nan
            residual_masked, threshold_lower, threshold_upper = sigma_clip(
                data=residual, sigma_lower=sigma_lower, sigma_upper=sigma_upper,
                maxiters=1, stdfunc='std', axis=None, masked=True, return_bounds=True, 
                grow=grow)
            master_mask = residual_masked.mask

    return p, residual, threshold_lower, threshold_upper, master_mask


def Spline1D(x, y, weight=None, mask=None, order=3, n_piece=1, bbox=[None, None], 
             n_iter=5, sigma_lower=None, sigma_upper=None, grow=False, 
             use_relative=False):
    """Fit 1-dimensional spline function.
    
    Knots are set equally spaced. Sigma clipping is used to mask bad data points.

    Parameters
    ----------
    x : array_like
        Input dimension of data points – must be increasing.

    y : array_like
        Input dimension of data points.

    weight : array_like, optional
        Weights for spline fitting. Must be positive. If `None` (default), weights are 
        all equal.

    mask : array_like, optional
        Initial mask for spline fitting. If `None` (default), data points are all 
        unmasked.

    order : int, optional
        Degree of the smoothing spline. Must be `5` >= ``order`` >= `1`.
        Default is `3`, a cubic spline.

    n_piece : int, optional
        Number of spline pieces. Lengths are all equal. Must be positive.
        Default is `1`.

    bbox : array_like, optional
        2-sequence specifying the boundary of the approximation interval. If `None` 
        (default), bbox = [x[0], x[-1]].

    n_iter : int, optional
        Number of sigma slipping iterations. Must be ``n_iter`` >= `0`.
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    grow : scalar or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits. 

    use_relative : bool, optional
        If `True`, relative residual is used in sigma slipping instead of residual. 
        Default is `False`.

    Returns
    -------
    spl : callable
        Callable object returned by `~scipy.interpolate.LSQUnivariateSpline`.

    residual : `~numpy.ndarray`
        Residual or relative residual, depending on ``use_relative``.

    threshold_lower : scalar or `None`
        Lower threshold for clipping.

    threshold_upper : scalar or `None`
        Upper threshold for in clipping.

    master_mask : `~numpy.ndarray`
        Final mask after clipping.
    """

    if weight is None:
        weight = np.ones_like(y)

    if mask is None:
        mask = np.zeros_like(y, dtype=bool)

    _validateInteger(n_piece, 'n_piece', (1, None), (True, None))

    _validateInteger(n_iter, 'n_iter', (0, None), (True, None))

    _validateBool(use_relative, 'use_relative')

    # Set ``n_piece`` equally spaced knots
    knots = x[~mask][0] \
            + np.arange(1, n_piece) * (x[~mask][-1] - x[~mask][0]) / n_piece

    # Spline fitting iteratively
    master_mask = mask & True
    # !!! only ``n_iter`` effective iterations !!!
    for k in range(n_iter + 1):
        spl = interpolate.LSQUnivariateSpline(
            x=x[~master_mask], y=y[~master_mask], t=knots, w=weight[~master_mask],
            bbox=bbox, k=order, ext='extrapolate', check_finite=False)

        y_fit = spl(x)

        # Residual
        residual = y - y_fit
        if use_relative:
            residual /= y_fit

        # Sigma clipping
        if n_iter == 0:
            threshold_lower, threshold_upper = None, None

        elif k < n_iter:
            residual[mask] = np.nan
            residual_masked, threshold_lower, threshold_upper = sigma_clip(
                data=residual, sigma_lower=sigma_lower, sigma_upper=sigma_upper,
                maxiters=1, stdfunc='std', axis=None, masked=True, return_bounds=True, 
                grow=grow)
            master_mask = residual_masked.mask

    return spl, residual, threshold_lower, threshold_upper, master_mask


def Spline2D(x, y, z, weight=None, mask=None, order=(3, 3), n_piece=(1, 1), 
             bbox=[None, None, None, None], n_iter=5, sigma_lower=None, 
             sigma_upper=None, axis=None, grow=False, use_relative=False):
    """Fit 2-dimensional spline function.
    
    Knots are set equally spaced. Sigma clipping is used to mask bad data points.

    Parameters
    ----------
    x : array_like
        Input dimension of data points – must be increasing.

    y : array_like
        Input dimension of data points.

    z : array_like
        Input dimension of data points.

    weight : array_like or `None`, optional
        Weights for spline fitting. Must be positive. If `None` (default), weights are 
        all equal.

    mask : array_like or `None`, optional
        Initial mask for spline fitting. If `None` (default), data points are all 
        unmasked.

    order : 2-tuple of int, optional
        Degree of the smoothing spline. Must be `5` >= ``order`` >= `1`.
        Default is (3, 3).

    n_piece : 2-tuple of int, optional
        Number of spline pieces. Lengths are all equal. Must be positive.
        Default is (1, 1).
        
    bbox : array_like, optional
        Sequence of length 4 specifying the boundary of the rectangular approximation 
        domain. By default, bbox=[x[0], x[-1], y[0], y[-1]].

    n_iter : int, optional
        Number of sigma slipping iterations. Must be ``n_iter`` >= `0`. 
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    axis : `None` or int or 2-tuple of int, optional
        The axis or axes along which to sigma clip the data. If `None`, then the 
        flattened data will be used. ``axis`` is passed into the ``cenfunc`` and 
        ``stdfunc``. 
        Default is `None`.

    grow : scalar or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits. (only applied along axis, if specified)

    use_relative : bool, optional
        If `True`, relative residual is used in sigma slipping instead of residual. 
        Default is `False`.

    Returns
    -------
    bispl : callable
        Callable object returned by `~scipy.interpolate.LSQBivariateSpline`.
    
    residual : `~numpy.ndarray`
        Residual or relative residual, depending on ``use_relative``.
    
    threshold_lower : scalar or array_like or None
        Lower threshold for clipping.
    
    threshold_upper : scalar or array_like or None
        Upper threshold for clipping.
    
    master_mask : `~numpy.ndarray`
        Final mask after clipping.
    """

    if weight is None:
        weight = np.ones_like(z)

    if mask is None:
        mask = np.zeros_like(z, dtype=bool)

    _validate1DArray(order, 'order', 2, True)

    _validate1DArray(n_piece, 'n_piece', 2, True)
    _validateInteger(n_piece[0], 'n_piece[0]', (1, None), (True, None))
    _validateInteger(n_piece[1], 'n_piece[1]', (1, None), (True, None))

    _validateInteger(n_iter, 'n_iter', (0, None), (True, None))

    _validateBool(use_relative, 'use_relative')

    # Order
    xorder, yorder = order

    # Set equally spaced knots
    xpiece, ypiece = n_piece
    xknots = x[~mask][0] \
             + np.arange(1, xpiece) * (x[~mask][-1] - x[~mask][0]) / xpiece
    yknots = y[~mask][0] \
             + np.arange(1, ypiece) * (y[~mask][-1] - y[~mask][0]) / ypiece
    
    # 2-dimensional spline fitting iteratively (!!! only ``n_iter`` effective 
    # iterations !!!)
    master_mask = mask & True
    for k in range(n_iter + 1):
        bispl = interpolate.LSQBivariateSpline(
            x=x[~master_mask], y=y[~master_mask], z=z[~master_mask], tx=xknots, 
            ty=yknots, w=weight[~master_mask], bbox=bbox, kx=xorder, ky=yorder, 
            eps=None)

        z_fit = bispl(x.flatten(), y.flatten(), grid=False).reshape(x.shape)

        # Residual
        residual = z - z_fit
        if use_relative:
            residual /= z_fit

        # Sigma clipping
        if n_iter == 0:
            threshold_lower, threshold_upper = None, None

        elif k < n_iter:
            residual[mask] = np.nan
            residual_masked, threshold_lower, threshold_upper = sigma_clip(
                data=residual, sigma_lower=sigma_lower, sigma_upper=sigma_upper,
                maxiters=1, stdfunc='std', axis=axis, masked=True, return_bounds=True, 
                grow=grow)
            master_mask = residual_masked.mask

    return bispl, residual, threshold_lower, threshold_upper, master_mask


def GaussianSmoothing2D(x, y, z, sigma, mask=None, n_iter=5, sigma_lower=None, 
                        sigma_upper=None, axis=None, grow=False, use_relative=False):
    """2-dimensional Gaussian smoothing.

    Parameters
    ----------
    x : array_like
        Input dimension of data points.

    y : array_like
        Input dimension of data points.

    z : array_like
        Input dimension of data points.

    sigma : scalar or sequence of scalars
        Standard deviation for Gaussian kernel. The standard deviations of the Gaussian 
        filter are given for each axis as a sequence, or as a single number, in which 
        case it is equal for both axes.

    mask : array_like, optional
        Initial mask for Gaussian smoothing. If `None` (default), data points are all 
        unmasked.

    n_iter : int, optional
        Number of sigma slipping iterations. Must be ``n_iter`` >= `0`.
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    axis : `None` or int or 2-tuple of int, optional
        The axis or axes along which to sigma clip the data. If `None`, then the 
        flattened data will be used. ``axis`` is passed into the ``cenfunc`` and 
        ``stdfunc``. 
        Default is `None`.

    grow : scalar or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits. (only applied along axis, if specified)

    use_relative : bool, optional
        If `True`, relative residual is used in sigma slipping instead of residual. 
        Default is `False`.

    Returns
    -------
    z_smoothed : `~numpy.ndarray`
        Smoothed data.

    residual : `~numpy.ndarray`
        Residual or relative residual, depending on ``use_relative``.

    threshold_lower : scalar or array_like or None
        Lower threshold for clipping.
    
    threshold_upper : scalar or array_like or None
        Upper threshold for clipping.

    master_mask : `~numpy.ndarray`
        Final mask after clipping.
    """

    if mask is None:
        mask = np.zeros(z.shape, dtype=bool)

    _validateInteger(n_iter, 'n_iter', (0, None), (True, None))

    _validateBool(use_relative, 'use_relative')

    # Convolve with Gaussian kernel iteratively (!!! only ``n_iter`` effective 
    # iterations !!!)
    master_mask = mask & True
    for k in range(n_iter + 1):
        # Interpolate
        interpolated_z = interpolate.griddata(
            points=(x[~master_mask], y[~master_mask]), values=z[~master_mask],
            xi=(x.flatten(), y.flatten()), method='nearest').reshape(x.shape)

        # Smooth
        z_smoothed = gaussian_filter(
            input=interpolated_z, sigma=sigma, order=0, mode='reflect')

        # Residual
        residual = z - z_smoothed
        if use_relative:
            residual /= z_smoothed

        # Sigma clipping
        if n_iter == 0:
            threshold_lower, threshold_upper = None, None

        elif k < n_iter:
            residual[mask] = np.nan
            residual_masked, threshold_lower, threshold_upper = sigma_clip(
                data=residual, sigma_lower=sigma_lower, sigma_upper=sigma_upper,
                maxiters=1, stdfunc='std', axis=axis, masked=True, return_bounds=True,
                grow=grow)
            master_mask = residual_masked.mask

    return z_smoothed, residual, threshold_lower, threshold_upper, master_mask


import warnings

# NumPy
import numpy as np
# SciPy
from scipy.optimize import curve_fit, OptimizeWarning
# matplotlib
import matplotlib.pyplot as plt
# AstroPy
from astropy.nddata import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm
# photutils
from photutils import CircularAperture, RectangularAperture
from photutils.background import Background2D
from photutils.centroids import centroid_com
from photutils.detection import find_peaks
# drpy
from drpy import conf
from drpy.modeling.function import CircularGaussian
from drpy.plotting import _plot2d
from drpy.validate import (_validateString, _validateBool, _validateRange, 
                           _validateInteger, _validateCCD, _validatePath)

__all__ = ['getFWHM']

# todo: deal with NaNs in Gaussian fitting.
def getFWHM(ccd, box_size, n_sigma, aper_radius, saturation, n_peak, use_mask=False, 
            title='FWHM', show=conf.show, save=conf.save, path=conf.path, **kwargs):
    """Estimate mean FWHM of the sources in the input image.

    The background of the input image is first modeled by a 2-dimensional background 
    estimator, and then subtracted from the input image. Detect sources in the 
    background subtracted image. Fit circular Gaussian profiles to the detected 
    sources.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData` or `~numpy.ndarray`
        Input image.

    box_size : int or array_like
        The box size along each axis. If ``box_size`` is a scalar then a square box of 
        size ``box_size`` will be used. If ``box_size`` has two elements, they must be 
        in (ny, nx) order. (for details see https://photutils.readthedocs.io/en/stable/
        api/photutils.background.Background2D.html)

    n_sigma : scalar
        The detection threshold is defined as ``n_sigma`` * background RMS.

    aper_radius : int
        Radius of a source in the input image. It is recommended to set comparable to 
        or slightly larger than the FWHM.

    saturation : scalar
        Saturation level of pixels.

    n_peak : int
        Number of peaks. When number of peaks exceeds ``n_peak``, the peaks with the 
        highest peak intensities are used in determining FWHM.

    use_mask : bool, optional
        If `True` and a mask array is attributed to ``ccd``, the masked pixels are 
        ignored in the background estimation and source detection. 
        Default is `False`.

    Returns
    -------
    fwhm : scalar
        Mean FWHM of the detected sources.

    fwhm_err : scalar
        Standard deviation of FWHM.
    """

    _validateInteger(aper_radius, 'aper_radius', (1, None), (True, None))

    # Assumes that ``aper_radius`` ~ FWHM ~ 2.3σ. Thus, for a circular Gaussian 
    # profile, almost 100% of the energy are concentrated in the aperture.
    aper_size = 2 * aper_radius + 1

    _validateBool(use_mask, 'use_mask')

    nccd, data_arr, _, mask_arr = _validateCCD(ccd, 'ccd', False, use_mask, False)

    # Background estimation
    bkg = Background2D(data=data_arr, box_size=box_size, mask=mask_arr, **kwargs)
    bkg_value = bkg.background
    bkgrms_value = bkg.background_rms

    _validateRange(n_sigma, 'n_sigma', (0, None), (False, None))

    # Source detection
    threshold = n_sigma * bkgrms_value
    data_arr_bkgsb = data_arr - bkg_value
    peak_table = find_peaks(
        data=data_arr_bkgsb, threshold=threshold, box_size=aper_size, footprint=None, 
        mask=mask_arr, border_width=(2 * aper_size - 1), centroid_func=centroid_com, 
        error=None, wcs=None)
    peak_table.sort('peak_value')
    peak_table.reverse()

    _validateRange(saturation, 'saturation', (0, None), (False, None))

    # Mask saturated sources
    isSaturated = peak_table['peak_value'].data >= saturation

    # Cutout size = 4 * ``aper_size`` - 3 = 8 * ``aper_radius`` + 1
    size = 4 * aper_size - 3

    # Mask sources too close to each other
    x, y = peak_table['x_centroid'].data, peak_table['y_centroid'].data
    isClose = np.zeros_like(x, dtype=bool)
    for i in range(x.shape[0]):
        if not isClose[i]:
            for j in range(i + 1, x.shape[0]):
                dist_sq = (x[j] - x[i])**2 + (y[j] - y[i])**2
                if dist_sq < size**2 / 2:
                    isClose[i], isClose[j] = True, True
    
    peak_table = peak_table[(~isSaturated) & (~isClose)]

    _validateInteger(n_peak, 'n_peak', (1, None), (True, None))

    # Remove redundant peaks
    peak_table = peak_table[:n_peak]

    # Update peak number (may become smaller)
    n_peak = len(peak_table)

    # Centroid
    x, y = peak_table['x_centroid'].data, peak_table['y_centroid'].data

    # X & Y grids
    Y, X = np.mgrid[-(size // 2):(size // 2 + 1), -(size // 2):(size // 2 + 1)]

    xy = (X.flatten(), Y.flatten())

    xy_min, xy_max = xy[0][0], xy[0][-1]

    position_arr = np.zeros((n_peak, 2)) + np.nan
    fwhm_arr = np.zeros(n_peak) + np.nan
    cutout2d_arr = np.zeros((n_peak, size, size))

    with warnings.catch_warnings():

        warnings.simplefilter('error', OptimizeWarning)

        for i in range(n_peak):

            # Cutout
            cutout2d = Cutout2D(
                data=data_arr_bkgsb, position=(x[i], y[i]), size=size, mode='strict')
            cutout2d_arr[i] = cutout2d.data / cutout2d.data.max()

            initial_guess = (1, 0, 0, aper_radius, 0)

            try:
                popt, _ = curve_fit(
                    CircularGaussian, xy, cutout2d_arr[i].flatten(), p0=initial_guess)

                a, x0, y0, sigma, b = popt

                if (a > 0) & (xy_min < x0 < xy_max) & (xy_min < y0 < xy_max):
                    position_arr[i] = x0 + size // 2, y0 + size // 2
                    fwhm_arr[i] = sigma * gaussian_sigma_to_fwhm
                    cutout2d_arr[i] -= b

                else:
                    raise RuntimeError('No feature fitted in the given region.')

            except (RuntimeError, TypeError, OptimizeWarning): # raise exception here
                pass

    # Output
    fwhm, fwhm_err = np.nanmean(fwhm_arr), np.nanstd(fwhm_arr, ddof=1)

    _validateBool(show, 'show')

    _validateString(title, 'title')
    if title != 'FWHM':
        title = [f'{title} FWHM (source detection)', 
                 f'{title} FWHM (Gaussian2D fitting)']
    else:
        title = [f'FWHM (source detection)', f'FWHM (Gaussian2D fitting)']

    fig_path = _validatePath(save, path, title)

    if show | save:

        # Plot source detection
        fig = plt.figure(dpi=100)
        ax = fig.add_subplot(1, 1, 1)
        # Image
        extent = (
            0.5, data_arr_bkgsb.shape[1] + 0.5, 0.5, data_arr_bkgsb.shape[0] + 0.5
        )
        _plot2d(ax=ax, ccd=data_arr_bkgsb, cmap='Greys_r', extent=extent)
        (xmin, xmax), (ymin, ymax) = ax.get_xlim(), ax.get_ylim()
        # Mask
        ax.imshow(
            mask_arr, cmap='Greys', alpha=0.3 * mask_arr.astype(int), origin='lower', 
            extent=extent)
        # Aperture
        apertures = RectangularAperture(
            positions=np.transpose([x + 1, y + 1]), w=size, h=size, theta=0)
        apertures.plot(ax, color='r', lw=1)
        # Setting
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.set_title(title[0], fontsize=16)

        if save: plt.savefig(fig_path[0], dpi=100)

        if show: plt.show()

        plt.close()

        if (n_peak % 4) == 0:
            nrow = n_peak // 4
        else:
            nrow = n_peak // 4 + 1

        fig = plt.figure(figsize=(10, 2 * nrow), dpi=100)
        extent = (0.5, size + 0.5, 0.5, size + 0.5)
        for i in range(n_peak):
            ax = fig.add_subplot(nrow, 4, i + 1)
            # Image
            _plot2d(
                ax=ax, ccd=cutout2d_arr[i], cmap='Greys_r', extent=extent, cbar=False, 
                xlabel=None, ylabel=None)
            # Aperture
            apertures = CircularAperture(
                (position_arr[i][0] + 1, position_arr[i][1] + 1), r=(fwhm_arr[i] / 2))
            apertures.plot(ax, color='r', lw=1)
            # Centroid
            ax.plot(position_arr[i][0] + 1, position_arr[i][1] + 1, 'r+')
            # Settings
            ax.tick_params(
                which='major', direction='in', top=True, right=True, color='w', length=5, 
                width=1, labelsize=12)
            ax.annotate(
                'FWHM$=' + f'{fwhm_arr[i]:.2f}' + '\\,$px', xy=(0.05, 0.85), 
                xycoords='axes fraction', color='w', fontsize=12)
        fig.supxlabel('column', fontsize=16)
        fig.supylabel('row', fontsize=16)
        fig.suptitle(title[1], fontsize=16)
        fig.tight_layout()

        if save: plt.savefig(fig_path[1], dpi=100)

        if show: plt.show()

        plt.close()
    
    return fwhm, fwhm_err


import os, textwrap, warnings
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
import astropy.units as u
from astropy.time import Time
from astropy.nddata import CCDData
# ccdproc
from ccdproc import (trim_image, combine, subtract_bias, create_deviation, 
                     flat_correct, cosmicray_lacosmic, cosmicray_median)
# drpy
from drpy import conf
from drpy.utils import imstatistics
from drpy.validate import _validateBool, _validateString

# Configurations
unit_ccddata = u.Unit(conf.unit_ccddata)

__all__ = ['CCDDataList']


class CCDDataList:
    """A class for batch processing."""

    @classmethod
    def read(cls, file_list, hdu=1, unit=unit_ccddata):
        """Construct a `CCDDataList` object from a list of files.
        
        Parameters
        ----------
        file_list : array_like
            List of files.

        hdu : `int`, optional
            FITS extension from which a `CCDDataList` object should be initialized.
            Default is `1`.

        unit : `~astropy.units.Unit` or `str`, optional
            Unit of the image data.
            Default is `~astropy.units.adu`.

        Returns
        -------
        ccddatalist : `CCDDataList`
            Constructed `CCDDataList` object.
        """

        ccddatalist = list()
        for file_name in file_list:
            # Load from file
            ccd = CCDData.read(file_name, hdu=hdu, unit=unit)
            # Add or rewrite keyword `FILENAME` (used in `self.statistics`).
            ccd.header['FILENAME'] = os.path.split(file_name)[1]
            ccddatalist.append(ccd)

        ccddatalist = cls(ccddatalist)

        return ccddatalist


    def __init__(self, ccddatalist):
        """Construct a `CCDDataList` object.
        
        The size of each image should be the same. (This is not checked in the 
        current version.)

        Parameters
        ----------
        ccddatalist : Iterable
            List of `~astropy.nddata.CCDData` objects.
        """

        for ccd in ccddatalist:
            if not isinstance(ccd, CCDData):
                raise TypeError(
                    f'Invalid type `{type(ccd)}` for an item of `CCDDataList` object. '
                    '`~astropy.nddata.CCDData` is expected.')

        self._ccddatalist = deepcopy(ccddatalist)   


    def trim(self, row_range, col_range, **kwargs):
        """Trim images.

        Parameters
        ----------
        row_range : array_like
            Row range (python style).

        col_range : array_like
            Column range (python style).
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            Trimmed images.
        """
        
        ccddatalist = list()
        for ccd in self._ccddatalist:
            ccddatalist.append(
                trim_image(
                    ccd=ccd[row_range[0]:row_range[1], col_range[0]:col_range[1]], 
                    **kwargs
                )
            )
        
        ccddatalist = self.__class__(ccddatalist)
        
        return ccddatalist


    def combine(self, method, **kwargs):
        """Combine images.
        
        Parameters
        ----------
        method : `str`
            `average`, `median`, or `sum`.
        
        Returns
        -------
        combined_ccd : `~astropy.nddata.CCDData`
            Combined frame.
        """
        
        combined_ccd = combine(
            img_list=self._ccddatalist, method=method, **kwargs)

        combined_ccd.header = self._ccddatalist[0].header
        combined_ccd.header['NCOMBINE'] = (self.__len__(), 'number of frames combined')
        
        return combined_ccd


    def subtract(self, master):
        """Subtract ``master`` frame from the `CCDDataList` object.

        Parameters
        ----------
        master : `~astropy.nddata.CCDData`
            Image to be subtracted from the `CCDDataList` object. Usually a master bias.

        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with ``master`` subtracted.
        """

        if not isinstance(master, CCDData):
            raise TypeError(
                f'Invalid type `{type(master)}` for ``master``. '
                '`~astropy.nddata.CCDData` is expected.')

        keywords_dict = {
            'ZEROCOR': '{} A level of {} is subtracted'.format(
                Time.now().to_value('iso', subfmt='date_hm'), round(master.data.mean()))
        }

        ccddatalist = list()
        for ccd in self._ccddatalist:
            ccddatalist.append(
                subtract_bias(ccd=ccd, master=master, add_keyword=keywords_dict)
            )
        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist

    # todo: use min_value to get rid of negative values ???
    def divide(self, master, min_value=None, norm_value=None):
        """Divide ``master`` from the `CCDDataList` object.

        Parameters
        ----------
        master : `~astropy.nddata.CCDData`
            Image to be divided from the `CCDDataList` object. Usually a master flat.
        
        min_value : float or `None`, optional
            Minimum value for flat field. The value can either be None and no minimum 
            value is applied to the flat or specified by a float which will replace all 
            values in the flat by the min_value. 
            Default is `None`.

        norm_value : float or `None`, optional
            If not `None`, normalize flat field by this argument rather than the mean 
            of the image. This allows fixing several different flat fields to have the 
            same scale. If this value is negative or 0, a `ValueError` is raised. 
            Default is `None`.

        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with ``master`` divided.
        """

        if not isinstance(master, CCDData):
            raise TypeError(
                f'Invalid type `{type(master)}` for ``master``. '
                '`~astropy.nddata.CCDData` is expected.')

        keywords_dict = {
            'FLATCOR': '{} flat-fielded'.format(
                Time.now().to_value('iso', subfmt='date_hm'))
        }

        ccddatalist = list()
        for ccd in self._ccddatalist:
            ccddatalist.append(
                flat_correct(
                    ccd=ccd, flat=master, min_value=min_value, norm_value=norm_value, 
                    add_keyword=keywords_dict)
            )
        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def create_deviation(self, gain, readnoise, **kwargs):
        """Create deviation.

        Parameters
        ----------
        gain : scalar
            Gain in [e- / adu].

        noise : scalar
            Read noise in [e-].

        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with uncertainty assigned.
        """

        ccddatalist = list()
        for ccd in self._ccddatalist:
            # the pre-assigned uncertainties will be replaced by new ones 
            ccddatalist.append(
                create_deviation(
                    ccd_data=ccd, gain=(gain * u.electron / u.adu), 
                    readnoise=(readnoise * u.electron), **kwargs)
            )
        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def cosmicray(self, method, use_mask=False, gain=None, readnoise=None, 
                  gain_apply=False, **kwargs):
        """Remove cosmic rays.
        
        Parameters
        ----------
        method : str
            `Laplacian` or `median`. Cosmic ray removal method.

        use_mask : bool, optional
            If `True`, the attributed mask frames are used in identifying cosmic ray 
            pixels. (only useful for `Laplacian` method)
            Default is `False`.

        gain : scalar or `None`, optional
            Gain in [e- / adu].

        readnoise : scalar or `None`, optional
            Read noise in [e-].

        gain_apply : bool, optional
            If `True`, return gain-corrected data, with correct units, otherwise do not 
            gain-correct the data. `False` is recommended.

        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with cosmic rays removed.
        """

        _validateBool(use_mask, 'use_mask')

        _validateString(method, 'method', ['Laplacian', 'median'])

        if method == 'Laplacian':
            keywords_dict = {
                'COSMICRA': '{} Laplacian Edge Detection'.format(
                    Time.now().to_value('iso', subfmt='date_hm'))
            }
        elif method == 'median':
            keywords_dict = {
                'COSMICRA': '{} median with sigma clipping'.format(
                    Time.now().to_value('iso', subfmt='date_hm'))
            }

        ccddatalist = list()
        for ccd in self._ccddatalist:

            nccd = ccd.copy()

            if use_mask:
                if nccd.mask is None:
                    warnings.warn(
                        'The input mask is unavailable. All set unmasked.', 
                        RuntimeWarning)
                    nccd.mask = np.zeros_like(nccd.data, dtype=bool)
            else:
                nccd.mask = None
            
            if method == 'Laplacian':

                ccd_cosmicray = cosmicray_lacosmic(
                    ccd=nccd, gain=gain, readnoise=readnoise, gain_apply=gain_apply, 
                    **kwargs)

            elif method == 'median':

                ccd_cosmicray = cosmicray_median(ccd=nccd, **kwargs)

            ccd_cosmicray.unit = nccd.unit

            for key, value in keywords_dict.items():
                ccd_cosmicray.header[key] = value

            ccddatalist.append(ccd_cosmicray)

        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def statistics(self, verbose=False):
        """Statistics.

        Parameters
        ----------
        verbose : bool, optional
            If `True`, statistics table is printted before return.
            Default is `False`.

        Returns
        -------
        out : `~astropy.table.table.Table`
            Statistics table.
        """

        return imstatistics(self._ccddatalist, verbose=verbose)


    def to_list(self):
        """Convert to `list` object."""
        return self._ccddatalist


    def copy(self):
        """Return a copy of itself."""
        return self.__class__(self._ccddatalist)


    def __getitem__(self, item):

        if isinstance(item, (int, np.integer)):
            return self._ccddatalist[item]

        elif isinstance(item, slice):
            return self.__class__(self._ccddatalist[item])

        elif isinstance(item, (np.ndarray, list)):

            if isinstance(item, list):
                item = np.array(item)

            if item.ndim > 1:
                raise IndexError(
                    f'`CCDDataList` is 1-dimensional, but {item.ndim} were indexed.')

            else:
                if item.dtype == bool:

                    if item.size != self.__len__():
                        raise IndexError(
                            'Boolean index did not match the size of `CCDDataList`.')

                    else:
                        item = np.where(item)[0]

                ccddatalist = list()
                for i in item:
                    ccddatalist.append(self._ccddatalist[i])

                return self.__class__(ccddatalist)

        else:
            raise IndexError(
                f'Invalid type {type(item)} for `CCDDataList` item access.')


    def __sub__(self, master):
        """The same as `self.subtract`."""
        return self.subtract(master)


    def __truediv__(self, master):
        """The same as `self.divide`."""
        return self.divide(master)


    def __repr__(self):

        if len(self._ccddatalist) == 0:
            core_string = ''

        elif len(self._ccddatalist) == 1:
            core_string = f'{self._ccddatalist[0]}'

        elif len(self._ccddatalist) == 2:
            core_string = f'{self._ccddatalist[0]},\n{self._ccddatalist[-1]}'

        elif len(self._ccddatalist) > 2:
            core_string = f'{self._ccddatalist[0]},\n...,\n{self._ccddatalist[-1]}'

        return (
            'CCDDataList([' + textwrap.indent(core_string, ' ' * 13)[13:] + '], length='
            f'{self.__len__()})'
        )


    def __len__(self):
        return len(self._ccddatalist)


    def __iter__(self):
        return iter(self._ccddatalist)


# NumPy
import numpy as np
# Scipy
from scipy import interpolate
# AstroPy
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.modeling.models import Gaussian2D
# photutils
from photutils.aperture import (CircularAperture, RectangularAperture, 
                                aperture_photometry)
# drpy
from drpy.validate import _validateString, _validateNDArray

__all__ = ['invertCoordinateMap', 'getSlitLoss']


def invertCoordinateMap(slit_along, coordinate):
    """Invert coordinate map.

    U(X, Y) and V(X, Y) -> X(U, V) and Y(U, V).

    Parameters
    ----------
    slit_along : str
        `col` or `row`, and 
        - if `col`, ``V`` = ``Y`` is assumed.
        - if `row`, ``U`` = ``X`` is assumed.

    coordinate : `~numpy.ndarray`
        One of the coordinate maps to be inverted (i.e., ``U`` or ``V``). A one-to-one 
        mapping is assumed for the other.

    Returns
    -------
    X, Y : `~numpy.ndarray`
        Inverted coordinate maps.
    """

    _validateString(slit_along, 'slit_along', ['col', 'row'])

    if slit_along == 'col':

        U = _validateNDArray(coordinate, 'coordinate', 2)

        n_row, n_col = U.shape

        idx_row, idx_col = np.arange(n_row), np.arange(n_col)

        Y = np.tile(np.arange(n_row), (n_col, 1)).T

        X = np.zeros((n_row, n_col))
        for i in idx_row:
            X[i] = interpolate.interp1d(
                x=U[i], y=idx_col, bounds_error=False, fill_value='extrapolate', 
                assume_sorted=True)(idx_col)

    else:

        V = _validateNDArray(coordinate, 'coordinate', 2)

        n_row, n_col = V.shape

        idx_row, idx_col = np.arange(n_row), np.arange(n_col)

        X = np.tile(np.arange(n_col), (n_row, 1))

        Y = np.zeros((n_row, n_col))
        for i in idx_col:
            Y[:, i] = interpolate.interp1d(
                x=V[:, i], y=idx_row, bounds_error=False, fill_value='extrapolate', 
                assume_sorted=True)(idx_row)
    
    return X, Y


def getSlitLoss(fwhm, slit_width):
    """Calculate slit loss.
    
    Parameters
    ----------
    fwhm: scalar
        Seeing.
    
    slit_width: scalar
        Slit width.
    
    Returns
    -------
    slit_loss: scalar
        Slit loss.
    """
    
    sigma = fwhm * gaussian_fwhm_to_sigma
    amplitude = (1 / (2 * np.pi * sigma**2))
    
    gaussian2d = Gaussian2D(
        amplitude=amplitude, x_mean=0, y_mean=0, x_stddev=sigma, y_stddev=sigma, 
        theta=0)
    
    ny = nx = np.int64(10 * sigma)
    Y, X = np.mgrid[-ny:ny+1, -nx:nx+1]
    data = gaussian2d(X, Y)
    
    (y, x) = np.unravel_index(np.argmax(data, axis=None), data.shape)
    
    # Aperture photometry
    rect_aper = RectangularAperture((x, y), w=slit_width, h=data.shape[0])
    flux_slit = aperture_photometry(data, rect_aper, method='exact')['aperture_sum'][0]
    
    # circ_aper = CircularAperture((x, y), r=data.shape[0] / 2)
    # flux_total = aperture_photometry(data, circ_aper, method='exact')['aperture_sum'][0]
    
    slit_loss = 1 - flux_slit
    
    # fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    # ax.imshow(data, )
    # rect_aper.plot(ax=ax, color='red', lw=2)
    # circ_aper.plot(ax=ax, color='red', lw=2)
    # plt.show()
    
    return slit_loss


import warnings
from copy import deepcopy

# AstroPy
from astropy.time import Time
# drpy
from drpy.validate import _validate1DArray, _validateCCDList

__all__ = ['concatenate']


def concatenate(ccdlist, row_range, col_range, scale=None):
    """Concatenate two frames.

    The region defined by ``row_range`` and ``col_range`` in the second frame is 
    replaced by that in the first frame. Usually this is used to concatenate 
    flat-fields or arc spectra of different exposure times. To concatenate multiple 
    frames, call this function repeatedly.

    Parameters
    ----------
    ccdlist : Iterable
        Iterable object containing two 2-dimensional frames.

    row_range : array_like
        Row range (python style).

    col_range : array_like
        Column range (python style).
    
    scale : scalar, array-like or `None`, optional
        Scaling factor. Frames are multiplied by scaling factor prior to concatenation.
        Default is `None`.

    Returns
    -------
    nccd : `~astropy.nddata.CCDData`
        The concatenated frame.
    """

    ccdlist = _validateCCDList(ccdlist, 2, 2)
    n_image = len(ccdlist)
    if n_image > 2:
        warnings.warn(
            '``ccdlist`` contains more than two frames. Extra frames will be ignored.', 
            RuntimeWarning)

    row_start, row_end = row_range
    col_start, col_end = col_range

    if scale is None:
        scale = _validate1DArray(1, 'scale', 2, True)

    else:
        scale = _validate1DArray(scale, 'scale', 2, True)

    data_arr = ccdlist[1].data * scale[1]
    data_arr[row_start:row_end, col_start:col_end] = (
        ccdlist[0].data[row_start:row_end, col_start:col_end] * scale[0]
    )

    if (ccdlist[0].uncertainty is None) | (ccdlist[1].uncertainty is None):
        uncertainty_arr = None

    else:
        uncertainty_arr = scale[1] * ccdlist[1].uncertainty.array
        uncertainty_arr[row_start:row_end, col_start:col_end] = (
            ccdlist[0].uncertainty.array[row_start:row_end, col_start:col_end] 
            * scale[0]
        )

    if (ccdlist[0].mask is None) | (ccdlist[1].mask is None):
        mask_arr = None

    else:
        mask_arr = deepcopy(ccdlist[1].mask)
        mask_arr[row_start:row_end, col_start:col_end] = (
            ccdlist[0].mask[row_start:row_end, col_start:col_end])

    nccd = ccdlist[0].copy()

    nccd.data = data_arr

    if uncertainty_arr is None:
        nccd.uncertainty = None

    else:
        # Here ``nccd.uncertainty`` is definitely not `None`.
        nccd.uncertainty.array = uncertainty_arr

    nccd.mask = mask_arr

    nccd.header['CONCATEN'] = (
        '{} Data section [{}:{}, {}:{}] is from the first frame.'.format(
            Time.now().to_value('iso', subfmt='date_hm'), (col_start + 1), col_end, 
            (row_start + 1), row_end)
    )

    return nccd


import os
from glob import glob

# NumPy
import numpy as np
# AstroPy
from astropy.io import fits
from astropy.time import Time
from astropy.table import Table
import astropy.units as u
from astropy.nddata import StdDevUncertainty
# specutils
from specutils import Spectrum1D
# drpy
from drpy.validate import _validateSpectrum1D

__all__ = ['saveSpectrum1D', 'loadSpectrum1D', 'loadStandardSpectrum', 
           'loadExtinctionCurve']


def _Spectrum1D_to_hdu(spectrum1d, header):
    """Convert a `~specutils.Spectrum1D` object to hdu."""

    # spectral_axis
    unit = spectrum1d.spectral_axis.unit.to_string('fits')
    array = spectrum1d.spectral_axis.value
    spectral_axis = fits.Column(
        name='spectral_axis', format='D', unit=unit, array=array)
    
    # flux
    unit = spectrum1d.flux.unit.to_string('fits')
    array = spectrum1d.flux.value

    if array.ndim <= 1:
        format = 'D'
        dim = None

    else:
        format = f'{np.prod(array.shape[:-1])}D'
        dim = '('
        for i in array.shape[:-1]:
            dim += f'{i},'
        dim = dim[:-1] + ')'

    flux = fits.Column(
        name='flux', format=format, unit=unit, dim=dim, array=array.T)

    columns = [spectral_axis, flux]

    # uncertainty
    if spectrum1d.uncertainty is not None:
        array = spectrum1d.uncertainty.array
        uncertainty = fits.Column(
            name='uncertainty', format=format, unit=unit, dim=dim, array=array.T)
        columns.append(uncertainty)

    # mask
    if spectrum1d.mask is not None:
        array = spectrum1d.mask
        mask = fits.Column(
            name='mask', format='L', dim=dim, array=array.T)
        columns.append(mask)

    hdu = fits.BinTableHDU.from_columns(columns, header)

    hdu.header['EXTNAME'] = ('spec', 'name of the extension')
    hdu.header['MODIFIED'] = (
        '{}'.format(Time.now().to_value('iso', subfmt='date_hm')), 'last modified')

    return hdu


def saveSpectrum1D(file_name, spectrum1d, overwrite=False):
    """Save a `~specutils.Spectrum1D` object to file.
    
    Parameters
    ----------
    file_name : str
        File name.
    
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        The spectrum to be saved.
    """

    new_spectrum1d = _validateSpectrum1D(spectrum1d, 'spectrum1d')
    
    # header
    if 'header' in new_spectrum1d.meta:
        header = new_spectrum1d.meta['header']
    else:
        header = None

    hdu = _Spectrum1D_to_hdu(new_spectrum1d, header)

    hdulist = fits.HDUList([fits.PrimaryHDU(), hdu])

    hdulist.writeto(file_name, overwrite=overwrite)

    return None

# todo: Table -> QTable ???
def loadSpectrum1D(file_name, ext=1, **kwargs):
    """Load tabular spectrum to `~specutils.Spectrum1D`.
    
    Parameters
    ----------
    file_name : str
        File name.

    ext : int or str, optional
        Extension.
    
    Returns
    -------
    spectrum1d : `~specutils.Spectrum1D`
        The loaded spectrum.
    """

    with fits.open(file_name, **kwargs) as hdulist:
        hdu = hdulist[ext]
        header = hdu.header
        table = Table.read(hdu)

    spectral_axis = table['spectral_axis'].value * table['spectral_axis'].unit

    flux = table['flux'].value * table['flux'].unit

    colnames = table.colnames

    if 'uncertainty' in colnames:
        uncertainty = StdDevUncertainty(table['uncertainty'].value)

    else:
        uncertainty = None
    
    if 'mask' in colnames:
        mask = table['mask'].value

    else:
        mask = None

    meta = {'header': header}
    
    spectrum1d = Spectrum1D(
        spectral_axis=spectral_axis, flux=flux, uncertainty=uncertainty, mask=mask, 
        meta=meta)

    return spectrum1d


def _AB_to_flux(wavelength, AB):
    """Convert AB magnitude to flux in [erg/cm2/s/A].

    Parameters
    ----------
    wavelength : `numpy.ndarray` or scalar
        wavelengths in [A].

    AB : `numpy.ndarray` or scalar
        AB magnitudes.

    returns
    -------
    flux : `numpy.ndarray`
        Flux in [erg/cm2/s/A].
    """

    c = 2.99792458e+10 # [cm/s]

    flux = 10**(-0.4 * AB) * 3631e-23 * c / wavelength**2 * 1e8 # [erg/cm2/s/A]

    return flux


def loadStandardSpectrum(standard):
    """Load standard spectrum.
    
    Parameters
    ----------
    standard : str
        Path to the file.

    Returns
    -------
    spectrum1d : `~specutils.Spectrum1D`
        Standard spectrum.
    
    bandpass : `~astropy.units.Quantity` or None
        Bandpass.
    """

    dir_library = os.path.realpath(
        os.path.join(
            os.path.split(os.path.realpath(__file__))[0], 'lib/onedstds'
        )
    )

    results = sorted(glob(os.path.join(dir_library, standard)))
    
    print(os.path.join(dir_library, standard))
    
    # All the standard spectrum should be named *.dat
    results = [result for result in results if result.endswith('.dat')]

    if not results:
        raise ValueError('No standard spectrum found.')

    elif len(results) > 1:
        raise ValueError(f'More than one standard spectrum found: {results}')

    else:
        path_to_standard = results[0]

        standard = os.path.split(path_to_standard)[1]

        library = os.path.split(os.path.split(path_to_standard)[0])[1]

        # todo: add libraries
        if library in ['blackbody', 'ctio']:
            raise ValueError('`blackbody` and `ctio` are not supported yet.')

        # Here no line is skipped because the first line starts with ``#``.
        spectrum = np.loadtxt(path_to_standard).T

        if library in ['bstdscal', 'ctiocal', 'ctionewcal', 'iidscal', 'irscal', 
                       'oke1990', 'redcal', 'spec16cal', 'spec50cal', 'spechayescal']:
            wavelength, AB, bandpass = spectrum
            flux = _AB_to_flux(wavelength, AB) * (u.erg / u.cm**2 / u.s /u.AA)
            bandpass = bandpass * u.AA

        # No bin for library `hststan`
        elif library == 'hststan':
            if standard.startswith('m'):
                wavelength, AB = spectrum
                flux = _AB_to_flux(wavelength, AB) * (u.erg / u.cm**2 / u.s /u.AA)

            else:
                # The third column is also flux but in [mJy]
                wavelength, flux, _ = spectrum
                flux = flux * 1e-16 * (u.erg / u.cm**2 / u.s /u.AA)

            bandpass = None

        elif library == 'okestan':
            if standard.startswith('m'):
                wavelength, AB, bandpass = spectrum
                flux = _AB_to_flux(wavelength, AB) * (u.erg / u.cm**2 / u.s /u.AA)

            else:
                wavelength, flux, _, bandpass = spectrum
                flux = flux * 1e-16 * (u.erg / u.cm**2 / u.s /u.AA)

            bandpass = bandpass * u.AA

        wavelength = wavelength * u.AA
        
        meta = {
            'header': {
                'FILENAME': standard
            }
        }

        spectrum1d = Spectrum1D(spectral_axis=wavelength, flux=flux, meta=meta)

        return spectrum1d, bandpass


def loadExtinctionCurve(extinct):
    """Load extinction curve.
    
    Parameters
    ----------
    extinct : str
        Path to the file. The current directory is searched prior to the library 
        directory.

    Returns
    -------
    spectrum1d : `~specutils.Spectrum1D`
        Extinction curve.
    """

    dir_current = os.getcwd()
    
    path_to_extinct = os.path.join(dir_current, extinct)

    try:

        wavelength, extinction = np.loadtxt(path_to_extinct).T

    except:

        dir_library = os.path.realpath(
            os.path.join(
                os.path.split(os.path.realpath(__file__))[0], 'lib/extinction'
            )
        )

        results = sorted(glob(os.path.join(dir_library, extinct)))

        # All the extinction curve should be named *.dat
        results = [result for result in results if result.endswith('.dat')]

        if not results:
            raise ValueError('No extinction curve found.')

        elif len(results) > 1:
            raise ValueError(f'More than one extinction curve found: {results}')

        else:
            path_to_extinct = results[0]

            wavelength, extinction = np.loadtxt(path_to_extinct).T
    
    finally:

        wavelength = wavelength * u.AA

        extinction = u.Quantity(extinction)

        meta = {
            'header': {
                'FILENAME': os.path.split(path_to_extinct)[1]
            }
        }

        spectrum1d = Spectrum1D(spectral_axis=wavelength, flux=extinction, meta=meta)

        return spectrum1d


# NumPy
import numpy as np
# AstroPy
from astropy.table import Table
from astropy.nddata import CCDData

from .validate import (_validateBool, _validateCCDData, _validateCCDList)

__all__ = ['imstatistics']


def imstatistics(ccdlist, verbose=False):
    """Image statistics.
    
    Parameters
    ----------
    ccdlist : Iterable or `~astropy.nddata.CCDData` or `~numpy.ndarray`
        A 2-dimensional image or an iterable object consist of 2-dimensional 
        images, the statistics of which are calculated.

    verbose : bool, optional
        If `True`, statistics table is printted before return.
        Default is `False`.
    
    Returns
    -------
    stats : `~astropy.table.table.Table`
        Statistics table.
    """

    if not isinstance(ccdlist, (np.ndarray, CCDData)):
        nccdlist = _validateCCDList(ccdlist, 1, 2) # dimension is set to 2
    else:
        nccdlist = [_validateCCDData(ccdlist, 'ccd')]
    
    n_image = len(nccdlist)

    # Construct table
    rows = list()
    for i in range(n_image):
        
        ccd = nccdlist[i]
        
        if 'FILENAME' in ccd.header:
            file_name = ccd.header['FILENAME']
        else:
            file_name = f'image {i + 1}'

        # !!! mask is effective here !!!
        rows.append(
            [file_name, ccd.shape, np.mean(ccd.data), np.std(ccd.data, ddof=1), 
             np.min(ccd.data), np.max(ccd.data)])

    stats = Table(rows=rows, names=['IMAGE', 'SHAPE', 'MEAN', 'STDDEV', 'MIN', 'MAX'])
    
    # Format
    for colname in ['MEAN', 'STDDEV', 'MIN', 'MAX']:
        stats[colname].format = "g"

    _validateBool(verbose, 'verbose')

    if verbose: stats.pprint_all()

    return stats


"""
drpy.onedspec: a toolkit to reduce 1-dimensional spectroscopy data.
"""

from .core import dispcor, sensfunc, calibrate1d
from .io import (saveSpectrum1D, loadSpectrum1D, loadStandardSpectrum, 
                 loadExtinctionCurve)

__all__ = ['dispcor', 'sensfunc', 'calibrate1d', 'saveSpectrum1D', 'loadSpectrum1D', 
           'loadStandardSpectrum', 'loadExtinctionCurve']


"""
drpy: a data reduction toolkit for astronomical photometry and spectroscopy
"""

import os

from astropy import config as _config

__all__ = ['__version__', 'conf']

__version__ = '0.0.1.5'


class Conf(_config.ConfigNamespace):
    """Configuration parameters for `drpy`."""

    # Unit
    unit_ccddata = _config.ConfigItem(
        'adu', 'Default unit for `~astropy.nddata.CCDData` objects.', cfgtype='string')

    unit_spectral_axis = _config.ConfigItem(
        'pixel', 'Default unit for `~specutils.Spectrum1D` objects.', cfgtype='string')

    unit_flux = _config.ConfigItem(
        'count', 'Default unit for `~specutils.Spectrum1D` objects.', cfgtype='string')

    # Plot
    show = _config.ConfigItem(False, 'Whether to show plots.', cfgtype='boolean')

    save = _config.ConfigItem(False, 'Whether to save plots.', cfgtype='boolean')

    path = _config.ConfigItem(os.getcwd(), 'Where to save plots.', cfgtype='string')


conf = Conf()

del _config


"""
drpy.twodspec: a toolkit to reduce 2-dimensional spectroscopy data.
"""

from .longslit import (response, illumination, align, fitcoords, transform, trace, 
                       background, extract, calibrate2d)

__all__ = ['response', 'illumination', 'align', 'fitcoords', 'transform', 'trace', 
           'background', 'extract', 'calibrate2d']


import warnings
from copy import deepcopy

# NumPy
import numpy as np
# Scipy
from scipy import signal
from scipy.optimize import curve_fit, OptimizeWarning
# drpy
from drpy.modeling.function import Gaussian1D

__all__ = ['center1D', 'refinePeakBases', 'refinePeaks']


def _center1D_Gaussian(x, y, initial_guess, threshold):
    """One dimensional centering using Gaussian fitting."""

    popt, _ = curve_fit(Gaussian1D, x, y, p0=initial_guess)

    amplitude, center, spread = popt

    if (amplitude > threshold) & (x[0] < center < x[-1]):
        return center

    else:
        raise RuntimeError('No feature fitted in the given interval.')

# todo: add Bartlett method.
def _center1D_Bartlett():
    """One dimensional centering using Bartlett window."""

    pass

# todo: validate threshold. add absorption. add Bartlett method.
def center1D(x, y, method, initial_guess, threshold=0):
    """1-dimensional centering.
    
    Parameters
    ----------
    x, y : array_like
        Input data.
    
    method : str
        Centering method.
    
    initial_guess : array_like
        Initial guess for the feature, and
        - if `Gaussian`, it should be an 1-dimensional array consist of amplitude, 
        center, spread, and offset.

    threshold : scalar, optional
        Feature.
    
    Returns
    -------
    center : scalar
        Center of the feature.
    """

    # Add method here
    _validateString(method, 'method', ['Gaussian', 'Bartlett'])
    
    if method == 'Gaussian':

        initial_guess = _validate1DArray(initial_guess, 'initial_guess', 4, False)
        
        center = _center1D_Gaussian(x, y, initial_guess, threshold)
        
    elif method == 'Bartlett':

        pass

    return center


def _refinePeakBases(peaks, left_bases, right_bases, n_peak, copy):
    """Refine left and right bases of peaks."""

    if copy:
        left_bases, right_bases = deepcopy(left_bases), deepcopy(right_bases)

    # Refine left base by comparing it with the right base of the former peak
    for i in range(1, n_peak):
        if right_bases[i - 1] < peaks[i]:
            left_bases[i] = np.max([left_bases[i], right_bases[i - 1]])

    # Refine right base by comparing it with the left base of the latter peak
    for i in range(0, n_peak - 1):
        if peaks[i] < left_bases[i + 1]:
            right_bases[i] = np.min([right_bases[i], left_bases[i + 1]])

    return left_bases, right_bases


def refinePeakBases(peaks, left_bases, right_bases):
    """Refine left and right bases of peaks.
    
    Parameters
    ----------
    peaks : array_like
        Peak locations in pixels.

    left_bases : array_like
        Left bases of peaks.

    right_bases : array_like
        Right bases of peaks.

    Returns
    -------
    left_bases, right_bases: `numpy.ndarray`
        Refined left and right bases of peaks.
    """
    
    try:
        peaks = np.array(peaks)
    except:
        raise TypeError(f'Invalid type `{type(peaks)}` for ``peaks``.')

    try:
        left_bases = np.array(left_bases)
    except:
        raise TypeError(f'Invalid type `{type(left_bases)}` for ``left_bases``.')

    try:
        right_bases = np.array(right_bases)
    except:
        raise TypeError(f'Invalid type `{type(right_bases)}` for ``right_bases``.')

    if peaks.shape[0] == left_bases.shape[0] == right_bases.shape[0]:
        n_peak = peaks.shape[0]

    else:
        raise ValueError(
            'Lengths of ``peaks``, ``left_bases`` and ``right_bases`` are not equal.')

    if np.all((peaks - left_bases) > 0) & np.all((right_bases - peaks) > 0):
        return _refinePeakBases(peaks, left_bases, right_bases, n_peak, copy=False)

    else:
        raise ValueError(
            'Peaks should locate between corresponding left and right bases.')


def _refinePeaks(spectrum, peaks, heights, left_bases, right_bases, tolerance):
    """Refine peak locations."""

    widths = right_bases - left_bases

    left_bases = np.round(left_bases).astype(int)
    right_bases = np.round(right_bases).astype(int) + 1

    index = np.arange(spectrum.shape[0])

    with warnings.catch_warnings():

        warnings.simplefilter('error', OptimizeWarning)

        refined_index = list()
        refined_peaks = list()

        for i in range(peaks.shape[0]):

            initial_guess = heights[i], peaks[i], widths[i]

            x = index[left_bases[i]:right_bases[i]]
            y = spectrum[left_bases[i]:right_bases[i]]

            try:
                center = _center1D_Gaussian(x, y, initial_guess, 0)
                refined_index.append(i); refined_peaks.append(center)

            except (RuntimeError, TypeError, OptimizeWarning): # raise exception here
                pass

    refined_index = np.array(refined_index)
    refined_peaks = np.array(refined_peaks)

    # Mask peaks that are close from each other
    mask = np.isclose(refined_peaks[:-1], refined_peaks[1:], rtol=0, atol=tolerance)
    mask = np.hstack([False, mask])

    refined_index = refined_index[~mask]
    refined_peaks = refined_peaks[~mask]

    return refined_peaks, refined_index


def refinePeaks(spectrum, peaks, properties, tolerance=1):
    """Refine peak locations in a spectrum from a set of initial estimates. This 
    function attempts to fit a Gaussian to each peak in the provided list. It returns a 
    list of sub-pixel refined peaks. If two peaks are very close, they can be refined 
    to the same location. In this case only one of the peaks will be returned - i.e. 
    this function will return a unique set of peak locations.

    Parameters
    ----------
    spectrum : array_like
        Input spectrum (intensities).

    peaks : array_like
        Peak locations in pixels.

    properties : dict
        Peak properties dictionary returned by `~scipy.signal.find_peaks`. Keywords 
        `peak_heights`, `left_bases` and `right_bases` are needed in Gaussian fittings. 
        They will be recalculated if not available in the dictionary.
    
    tolerance : scalar, optional
        If the distance between two peaks < ``tolerance``, they are refined to the same 
        location.
    
    refine_bases : bool, optional
        Whether to refine left and right bases first.

    Returns
    -------
    refined_peaks: `numpy.ndarray`
        Refined peak locations in pixels.
    
    refined_properties: dict
        Properties dictionary of the refined peaks.
    
    refined_index : `numpy.ndarray`
        Index of the refined peaks in the input peak array.
    """

    try:
        spectrum = np.array(spectrum, dtype=float)
    except:
        raise TypeError(f'Invalid type `{type(spectrum)}` for ``spectrum``.')

    try:
        peaks = np.array(peaks, dtype=float)
    except:
        raise TypeError(f'Invalid type `{type(peaks)}` for ``peaks``.')

    if 'peak_heights' not in properties:
        heights = spectrum[np.round(peaks).astype(int)]
    else:
        try:
            heights = np.array(properties['peak_heights'], dtype=float)
        except:
            raise TypeError(
                f"Invalid type `{type(properties['peak_heights'])}` for ``heights``.")

    if ('left_bases' not in properties) or ('right_bases' not in properties):
        _, left_bases, right_bases = signal.peak_prominences(
            x=spectrum, peaks=np.round(peaks).astype(int), wlen=None)
    else:
        try:
            left_bases = np.array(properties['left_bases'], dtype=float)
        except:
            raise TypeError(
                f"Invalid type `{type(properties['left_bases'])}` for "
                "``left_bases``.")
        try:
            right_bases = np.array(properties['right_bases'], dtype=float)
        except:
            raise TypeError(
                f"Invalid type `{type(properties['right_bases'])}` for "
                "``right_bases``.")

    n_peak = peaks.shape[0]

    if not (heights.shape[0] == left_bases.shape[0] == right_bases.shape[0] == n_peak):
        raise ValueError(
            'Lengths of ``peaks``, ``heights``, ``left_bases`` and ``right_bases`` '
            'are not equal.')

    if not (np.all((peaks - left_bases) > 0) & np.all((right_bases - peaks) > 0)):
        raise ValueError(
            'Peaks should locate between corresponding left and right bases.')

    left_bases, right_bases = _refinePeakBases(
        peaks, left_bases, right_bases, n_peak=n_peak, copy=True)

    refined_peaks, refined_index = _refinePeaks(
        spectrum, peaks, heights, left_bases, right_bases, tolerance)

    refined_properties = deepcopy(properties)
    for key, val in refined_properties.items():
        refined_properties[key] = val[refined_index]
    refined_properties['left_bases'] = left_bases[refined_index]
    refined_properties['right_bases'] = right_bases[refined_index]

    return refined_peaks, refined_properties, refined_index